@@ -1,18 +1,6 @@
 /*************************************************
 Documentation of symbols defined by Math::MPFR
 
-NV_IS_LONG_DOUBLE        : Automatically defined by Makefile.PL if
-                           $Config{nvtype} is 'long double'.
-
-NV_IS_FLOAT128           : Automatically defined by Makefile.PL if
-                           $Config{nvtype} is __float128
-                           If NV_IS_FLOAT128 is defined we include the
-                           quadmath.h header.
-
-NV_IS_53_BIT             : Defined only when $Config{nvtype} is 'double' or
-                           when $Config{nvtype} is a 'long double' that's
-                           identical to the double.
-
 MPFR_WANT_FLOAT128       : Defined by Makefile.PL if $have_float128 is
                            set to a true value. $have_float128 can be set
                            to a true value by either editing the Makefile.PL
@@ -27,13 +15,13 @@ MPFR_WANT_FLOAT128       : Defined by Makefile.PL if $have_float128 is
                            MPFR_WANT_FLOAT128 must NOT be defined if the
                            mpfr library has NOT been built with __float128
                            support.
-                           MPFR_WANT_FLOAT128 does not imply that NV_IS_FLOAT128
-                           has been defined - perhaps we have defined
+                           MPFR_WANT_FLOAT128 does not imply that nvtype is
+                           __float128 - perhaps we have defined
                            MPFR_WANT_FLOAT128 solely because we wish to make
                            use of the Math::Float128-Math::MPFR interface.
 
 CAN_PASS_FLOAT128        : Defined only when both MPFR_WANT_FLOAT128 and
-                           NV_IS_FLOAT128 is defined, and then only if the mpfr
+                           USE_QUADMATH is defined, and then only if the mpfr
                            library is at version 4.0.0 or later. (There was no
                            __float128 support in the mpfr library prior to
                            4.0.0.)
@@ -173,10 +161,8 @@ FALLBACK_NOTIFY          : If defined, $Math::MPFR::doubletoa_fallback
 #include <stdio.h>
 
 #if defined(MATH_MPFR_NEED_LONG_LONG_INT)
-#ifndef _MSC_VER
 #include <inttypes.h>
 #endif
-#endif
 
 /*
  * In mpfr-4.1.0, the _Float128 type is exposed in mpfr.h if MPFR_WANT_FLOAT128 is defined.
@@ -191,14 +177,14 @@ FALLBACK_NOTIFY          : If defined, $Math::MPFR::doubletoa_fallback
 #include <mpfr.h>
 #include <float.h>
 
-#if defined(MPFR_WANT_FLOAT128) || defined(NV_IS_FLOAT128)
+#if defined(MPFR_WANT_FLOAT128) || defined(USE_QUADMATH)
 #include <quadmath.h>
-#if defined(NV_IS_FLOAT128) && defined(MPFR_WANT_FLOAT128) && defined(MPFR_VERSION) && MPFR_VERSION >= MPFR_VERSION_NUM(4,0,0)
+#if defined(USE_QUADMATH) && defined(MPFR_WANT_FLOAT128) && defined(MPFR_VERSION) && MPFR_VERSION >= MPFR_VERSION_NUM(4,0,0)
 #define CAN_PASS_FLOAT128
 #endif
 #if defined(__MINGW32__) && !defined(__MINGW64__)
 typedef __float128 float128 __attribute__ ((aligned(32)));
-#elif defined(__MINGW64__) || (defined(DEBUGGING) && defined(NV_IS_DOUBLE))
+#elif defined(__MINGW64__) || (defined(DEBUGGING) && NVSIZE == 8)
 typedef __float128 float128 __attribute__ ((aligned(8)));
 #else
 typedef __float128 float128;
@@ -206,13 +192,32 @@ typedef __float128 float128;
 #endif
 
 #if defined(MPFR_WANT_DECIMAL128)
-#if defined(__MINGW64__) || (defined(DEBUGGING) && defined(NV_IS_DOUBLE))
+#if defined(__MINGW64__) || (defined(DEBUGGING) && NVSIZE == 8)
 typedef _Decimal128 D128 __attribute__ ((aligned(8)));
 #else
 typedef _Decimal128 D128;
 #endif
 #endif
 
+#define SIS_PERL_VERSION PERL_REVISION*1000000+PERL_VERSION*1000+PERL_SUBVERSION
+
+#if SIS_PERL_VERSION >= 5012000   /* perl-5.12.0 and later */
+#  define MORTALIZED_PV(x) newSVpvn_flags(x,strlen(x),SVs_TEMP)
+#else
+#  define MORTALIZED_PV(x) sv_2mortal(newSVpv(x,0))
+#endif
+
+/* Facilitate altering the numeric flags that we *
+ * look at (in case it becomes necessary)        */
+
+#define SV_IS_IOK(x) \
+     SvIOK(x)
+
+#define SV_IS_POK(x) \
+     SvPOK(x)
+
+#define SV_IS_NOK(x) \
+     SvNOK(x)
 
 #if (!defined(MPFR_VERSION) || MPFR_VERSION <= 196868) && LDBL_MANT_DIG == 64
 #define LD_SUBNORMAL_BUG 1
@@ -224,19 +229,15 @@ typedef _Decimal128 D128;
 #define REQUIRED_LDBL_MANT_DIG LDBL_MANT_DIG
 #endif
 
-#if (!defined(NV_IS_FLOAT128) && !defined(NV_IS_LONG_DOUBLE)) || (defined(NV_IS_LONG_DOUBLE) && REQUIRED_LDBL_MANT_DIG == 53)
-#define NV_IS_53_BIT 1
-#endif
-
 #define FAILS_CHECK_INPUT_BASE \
-     SvIV(base) < 0 || SvIV(base) > 62 || SvIV(base) == 1
+     !SV_IS_IOK(base) || SvIVX(base) < 0 || SvIVX(base) > 62 || SvIVX(base) == 1
 
 #if MPFR_VERSION >= 262400 /* Allowable range of base has been expanded */
 #define FAILS_CHECK_OUTPUT_BASE \
-     SvIV(base) < -36 || SvIV(base) > 62 || abs(SvIV(base)) < 2
+     !(SV_IS_IOK(base) && ((SvIVX(base) >= 2 && SvIVX(base) <= 62) || (SvIVX(base) >= -36 && SvIVX(base) <= -2)))
 #else
 #define FAILS_CHECK_OUTPUT_BASE \
-     SvIV(base) < 2 || SvIV(base) > 62
+     !(SV_IS_IOK(base) && SvIVX(base) >= 2 && SvIVX(base) <= 62)
 #endif
 
 /* Don't use CHECK_ROUNDING_VALUE macro with Rmpfr_set_NV      *
@@ -252,7 +253,7 @@ typedef _Decimal128 D128;
 #endif
 
 #define NOK_POK_DUALVAR_CHECK \
-        if(SvNOK(b)) { \
+        if(SV_IS_NOK(b)) { \
          nok_pok++; \
          if(SvIV(get_sv("Math::MPFR::NOK_POK", 0))) \
            warn("Scalar passed to %s is both NV and PV. Using PV (string) value"
@@ -344,11 +345,11 @@ typedef _Decimal128 D128;
 #define DEAL_WITH_NANFLAG_BUG_OVERLOADED
 #endif
 
-/* Squash some annoying compiler warnings (Microsoft compilers only). */
-
+/*
 #ifdef _MSC_VER
 #pragma warning(disable:4700 4715 4716)
 #endif
+*/
 
 #ifdef OLDPERL
 #define SvUOK SvIsUV
@@ -362,6 +363,22 @@ typedef _Decimal128 D128;
 #  define Newxz(v,n,t) Newz(0,v,n,t)
 #endif
 
+/* A perl bug in perl-5.20 onwards can break &PL_sv_yes and  *
+ * &PL_sv_no. In the overload subs we therefore instead      *
+ * use  SvTRUE_nomg_NN where possible, which is available    *
+ * beginning with perl-5.18.0.                               *
+ * Otherwise we continue using &PL_sv_yes as original        *
+ * (&PL_sv_no is not used by this module.)                   *
+ * See See https://github.com/sisyphus/math-decimal64/pull/1 */
+
+#if defined SvTRUE_nomg_NN
+#define SWITCH_ARGS SvTRUE_nomg_NN(third)
+#define NO_SWITCH_ARGS !SvTRUE_nomg_NN(third)
+#else
+#define SWITCH_ARGS third==&PL_sv_yes
+#define NO_SWITCH_ARGS third!=&PL_sv_yes
+#endif
+
 /* May one day be removed from mpfr.h */
 #ifndef mp_rnd_t
 # define mp_rnd_t  mpfr_rnd_t
@@ -379,7 +396,13 @@ typedef _Decimal128 D128;
 #endif
 
 /* For nvtoa() */
-#if defined(NV_IS_53_BIT)
+#if defined(MPFR_HAVE_BENDIAN)			/* big endian architecture - defined by Makefile.PL */
+#define INC_OR_DEC(p) p++
+#else						/* little endian */
+#define INC_OR_DEC(p) p--
+#endif
+
+#if NVSIZE == 8
 #define MATH_MPFR_MAX_DIG 17
 #define NVSIZE_BITS 53
 #define MATH_MPFR_NV_MAX 1.7976931348623157e+308
@@ -387,21 +410,21 @@ typedef _Decimal128 D128;
 
 # if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
 
-# define D_CONDITION_1 i<=7
-# define D_INC_OR_DEC i++;
+# define D_CONDITION_1(p) p<=7
 # define DIND_0 0
 # define DIND_1 1
+# define DIND_2 2
 
 # else						/* little endian architecture */
 
-# define D_CONDITION_1 i>=0
-# define D_INC_OR_DEC i--;
+# define D_CONDITION_1(p) p>=0
 # define DIND_0 7
 # define DIND_1 6
+# define DIND_2 5
 
 # endif
 
-#elif defined(NV_IS_LONG_DOUBLE) && REQUIRED_LDBL_MANT_DIG == 64
+#elif defined(USE_LONG_DOUBLE)
 #define MATH_MPFR_MAX_DIG 21
 #define NVSIZE_BITS 64
 #define MATH_MPFR_NV_MAX 1.18973149535723176502e4932L
@@ -409,61 +432,26 @@ typedef _Decimal128 D128;
 
 # if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
 
-# define LD_CONDITION_1 i<=9
-# define LD_INC_OR_DEC i++;
+# define LD_CONDITION_1(p) p<=9
 # define LDIND_0 0
 # define LDIND_1 1
 # define LDIND_2 2
 
 # else						/* little endian architecture */
 
-# define LD_CONDITION_1 i>=0
-# define LD_INC_OR_DEC i--;
+# define LD_CONDITION_1(p) p>=0
 # define LDIND_0 9
 # define LDIND_1 8
 # define LDIND_2 7
 
 # endif
 
-#elif defined(NV_IS_LONG_DOUBLE) && REQUIRED_LDBL_MANT_DIG == 2098
+#elif defined(USE_LONG_DOUBLE) && REQUIRED_LDBL_MANT_DIG == 2098
 #define MATH_MPFR_MAX_DIG 33
 #define NVSIZE_BITS 2098
 #define MATH_MPFR_NV_MAX 1.797693134862315807937289714053e+308L
 #define MATH_MPFR_NORMAL_MIN 2.2250738585072014e-308
 
-# if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
-# define DD_CONDITION_1 i<=7
-# define DD_CONDITION_2 i=2;i<8;i++
-# define DD_INC_OR_DEC i++;
-# define IND_0 0
-# define IND_1 1
-# define LSD_BYTE_1 8
-# define LSD_BYTE_2 9
-# define LSD_BYTE_3 10
-# define LSD_BYTE_4 11
-# define LSD_BYTE_5 12
-# define LSD_BYTE_6 13
-# define LSD_BYTE_7 14
-# define LSD_BYTE_8 15
-
-# else                                        /* little endian architecture */
-
-# define DD_CONDITION_1 i>=0
-# define DD_CONDITION_2 i=13;i>7;i--
-# define DD_INC_OR_DEC i--;
-# define IND_0 15
-# define IND_1 14
-# define LSD_BYTE_1 7
-# define LSD_BYTE_2 6
-# define LSD_BYTE_3 5
-# define LSD_BYTE_4 4
-# define LSD_BYTE_5 3
-# define LSD_BYTE_6 2
-# define LSD_BYTE_7 1
-# define LSD_BYTE_8 0
-
-# endif /* MPFR_HAVE_BENDIAN */
-
 #else
 #define MATH_MPFR_MAX_DIG 36
 #define NVSIZE_BITS 113
@@ -472,16 +460,14 @@ typedef _Decimal128 D128;
 
 # if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
 
-# define Q_CONDITION_1 i<=15
-# define Q_INC_OR_DEC i++;
+# define Q_CONDITION_1(p) p<=15
 # define QIND_0 0
 # define QIND_1 1
 # define QIND_2 2
 
 # else						/* little endian architecture */
 
-# define Q_CONDITION_1 i>=0
-# define Q_INC_OR_DEC i--;
+# define Q_CONDITION_1(p) p>=0
 # define QIND_0 15
 # define QIND_1 14
 # define QIND_2 13
